"""Tests for the `benchmarker` module."""

import logging
import os
import time
from collections.abc import Generator
from dataclasses import replace
from pathlib import Path
from shutil import rmtree

import pytest
import torch
from requests.exceptions import RequestException

from euroeval.benchmarker import (
    Benchmarker,
    adjust_logging_level,
    clear_model_cache_fn,
    get_record,
    prepare_dataset_configs,
)
from euroeval.data_models import (
    BenchmarkConfig,
    BenchmarkResult,
    DatasetConfig,
    Language,
    ModelConfig,
    Task,
)
from euroeval.dataset_configs import get_dataset_config
from euroeval.enums import TaskGroup
from euroeval.exceptions import HuggingFaceHubDown


@pytest.fixture(scope="module")
def benchmarker() -> Generator[Benchmarker, None, None]:
    """A `Benchmarker` instance."""
    yield Benchmarker(progress_bar=False, save_results=False, num_iterations=1)


def test_benchmark_results_is_a_list(benchmarker: Benchmarker) -> None:
    """Test that the `benchmark_results` property is a list."""
    assert isinstance(benchmarker.benchmark_results, list)


@pytest.mark.depends(on=["tests/test_model_loading.py::test_load_non_generative_model"])
def test_benchmark_encoder(
    benchmarker: Benchmarker, task: Task, language: Language, encoder_model_id: str
) -> None:
    """Test that an encoder model can be benchmarked."""
    benchmark_result = None
    for _ in range(10):
        try:
            benchmark_result = benchmarker.benchmark(
                model=encoder_model_id, task=task.name, language=language.code
            )
            break
        except (HuggingFaceHubDown, RequestException, ConnectionError):
            time.sleep(5)
    else:
        pytest.skip(reason="Hugging Face Hub is down, so we skip this test.")
    assert isinstance(benchmark_result, list)
    assert all(isinstance(result, BenchmarkResult) for result in benchmark_result)


@pytest.mark.depends(on=["tests/test_model_loading.py::test_load_generative_model"])
def test_benchmark_generative(
    benchmarker: Benchmarker, task: Task, language: Language, generative_model_id: str
) -> None:
    """Test that a generative model can be benchmarked."""
    if not torch.cuda.is_available() and (
        task.task_group
        in [TaskGroup.SEQUENCE_CLASSIFICATION, TaskGroup.MULTIPLE_CHOICE_CLASSIFICATION]
        or task.uses_structured_output
    ):
        pytest.skip(
            f"CUDA is required to run generative models on the {task.name} "
            "task currently."
        )
    benchmark_result = benchmarker.benchmark(
        model=generative_model_id, task=task.name, language=language.code
    )
    assert isinstance(benchmark_result, list)
    assert all(isinstance(result, BenchmarkResult) for result in benchmark_result)


@pytest.mark.depends(on=["tests/test_model_loading.py::test_load_generative_model"])
def test_benchmark_generative_adapter(
    benchmarker: Benchmarker,
    task: Task,
    language: Language,
    generative_adapter_model_id: str,
) -> None:
    """Test that a generative adapter model can be benchmarked."""
    if not torch.cuda.is_available() and (
        task.task_group
        in [TaskGroup.SEQUENCE_CLASSIFICATION, TaskGroup.MULTIPLE_CHOICE_CLASSIFICATION]
        or task.uses_structured_output
    ):
        pytest.skip(
            f"CUDA is required to run generative models on the {task.name} "
            "task currently."
        )
    benchmark_result = benchmarker.benchmark(
        model=generative_adapter_model_id, task=task.name, language=language.code
    )
    assert isinstance(benchmark_result, list)
    assert all(isinstance(result, BenchmarkResult) for result in benchmark_result)


@pytest.mark.skipif(
    condition=os.getenv("OPENAI_API_KEY") is None,
    reason="OpenAI API key is not available.",
)
def test_benchmark_openai(
    benchmarker: Benchmarker, task: Task, language: Language, openai_model_id: str
) -> None:
    """Test that an OpenAI model can be benchmarked."""
    benchmark_result = benchmarker.benchmark(
        model=openai_model_id, task=task.name, language=language.code
    )
    assert isinstance(benchmark_result, list)
    assert all(isinstance(result, BenchmarkResult) for result in benchmark_result)


@pytest.mark.skipif(
    condition=os.system("uv run ollama -v") != 0, reason="Ollama is not available."
)
def test_benchmark_ollama(
    benchmarker: Benchmarker, task: Task, language: Language, ollama_model_id: str
) -> None:
    """Test that an Ollama model can be benchmarked."""
    benchmark_result = benchmarker.benchmark(
        model=ollama_model_id, task=task.name, language=language.code
    )
    assert isinstance(benchmark_result, list)
    assert all(isinstance(result, BenchmarkResult) for result in benchmark_result)


@pytest.mark.disable_socket
@pytest.mark.depends(on=["test_benchmark_encoder"])
def test_benchmark_encoder_no_internet(
    task: Task, language: Language, encoder_model_id: str
) -> None:
    """Test that encoder models can be benchmarked without internet."""
    # We need a new benchmarker since we only check for internet once per instance
    benchmarker = Benchmarker(progress_bar=False, save_results=False, num_iterations=1)
    benchmark_result = benchmarker.benchmark(
        model=encoder_model_id, task=task.name, language=language.code
    )
    assert isinstance(benchmark_result, list)
    assert all(isinstance(result, BenchmarkResult) for result in benchmark_result)


# Allow localhost since vllm uses it for some things
@pytest.mark.allow_hosts(["127.0.0.1"])
@pytest.mark.depends(on=["test_benchmark_generative"])
def test_benchmark_generative_no_internet(
    task: Task, language: Language, generative_model_id: str
) -> None:
    """Test that generative models can be benchmarked without internet."""
    if not torch.cuda.is_available() and (
        task.task_group
        in [TaskGroup.SEQUENCE_CLASSIFICATION, TaskGroup.MULTIPLE_CHOICE_CLASSIFICATION]
        or task.uses_structured_output
    ):
        pytest.skip(
            f"CUDA is required to run generative models on the {task.name} "
            "task currently."
        )
    # We need a new benchmarker since we only check for internet once per instance
    benchmarker = Benchmarker(progress_bar=False, save_results=False, num_iterations=1)
    benchmark_result = benchmarker.benchmark(
        model=generative_model_id, task=task.name, language=language.code
    )
    assert isinstance(benchmark_result, list)
    assert all(isinstance(result, BenchmarkResult) for result in benchmark_result)


# Allow localhost since vllm uses it for some things
@pytest.mark.allow_hosts(["127.0.0.1"])
@pytest.mark.skip(
    "Benchmarking adapter models without internet access are not implemented yet."
)
@pytest.mark.depends(on=["test_benchmark_generative_adapter"])
def test_benchmark_generative_adapter_no_internet(
    task: Task, language: Language, generative_adapter_model_id: str
) -> None:
    """Test that generative adapter models can be benchmarked without internet."""
    if not torch.cuda.is_available() and (
        task.task_group
        in [TaskGroup.SEQUENCE_CLASSIFICATION, TaskGroup.MULTIPLE_CHOICE_CLASSIFICATION]
        or task.uses_structured_output
    ):
        pytest.skip(
            f"CUDA is required to run generative models on the {task.name} "
            "task currently."
        )
    # We need a new benchmarker since we only check for internet once per instance
    benchmarker = Benchmarker(progress_bar=False, save_results=False, num_iterations=1)
    benchmark_result = benchmarker.benchmark(
        model=generative_adapter_model_id, task=task.name, language=language.code
    )
    assert isinstance(benchmark_result, list)
    assert all(isinstance(result, BenchmarkResult) for result in benchmark_result)


@pytest.mark.parametrize(
    argnames=["few_shot", "evaluate_test_split", "benchmark_results", "expected"],
    argvalues=[
        (False, True, [], False),
        (
            False,
            True,
            [
                BenchmarkResult(
                    model="model_id@revision",
                    dataset="dataset",
                    generative=False,
                    generative_type=None,
                    few_shot=False,
                    validation_split=False,
                    num_model_parameters=100,
                    max_sequence_length=100,
                    vocabulary_size=100,
                    merge=False,
                    languages=["da"],
                    task="task",
                    results=dict(),
                )
            ],
            True,
        ),
        (
            True,
            True,
            [
                BenchmarkResult(
                    model="model_id@revision",
                    dataset="another-dataset",
                    generative=False,
                    generative_type=None,
                    few_shot=False,
                    validation_split=False,
                    num_model_parameters=100,
                    max_sequence_length=100,
                    vocabulary_size=100,
                    merge=False,
                    languages=["da"],
                    task="task",
                    results=dict(),
                )
            ],
            False,
        ),
        (
            True,
            True,
            [
                BenchmarkResult(
                    model="model_id@revision",
                    dataset="dataset",
                    generative=True,
                    generative_type=None,
                    few_shot=False,
                    validation_split=False,
                    num_model_parameters=100,
                    max_sequence_length=100,
                    vocabulary_size=100,
                    merge=False,
                    languages=["da"],
                    task="task",
                    results=dict(),
                )
            ],
            False,
        ),
        (
            True,
            True,
            [
                BenchmarkResult(
                    model="model_id@revision",
                    dataset="dataset",
                    generative=True,
                    generative_type=None,
                    few_shot=True,
                    validation_split=False,
                    num_model_parameters=100,
                    max_sequence_length=100,
                    vocabulary_size=100,
                    merge=False,
                    languages=["da"],
                    task="task",
                    results=dict(),
                )
            ],
            True,
        ),
        (
            True,
            True,
            [
                BenchmarkResult(
                    model="model_id@revision",
                    dataset="dataset",
                    generative=False,
                    generative_type=None,
                    few_shot=False,
                    validation_split=False,
                    num_model_parameters=100,
                    max_sequence_length=100,
                    vocabulary_size=100,
                    merge=False,
                    languages=["da"],
                    task="task",
                    results=dict(),
                )
            ],
            True,
        ),
        (
            False,
            False,
            [
                BenchmarkResult(
                    model="model_id@revision",
                    dataset="dataset",
                    generative=False,
                    generative_type=None,
                    few_shot=False,
                    validation_split=False,
                    num_model_parameters=100,
                    max_sequence_length=100,
                    vocabulary_size=100,
                    merge=False,
                    languages=["da"],
                    task="task",
                    results=dict(),
                )
            ],
            False,
        ),
        (
            False,
            False,
            [
                BenchmarkResult(
                    model="model_id@revision",
                    dataset="dataset",
                    generative=False,
                    generative_type=None,
                    few_shot=False,
                    validation_split=True,
                    num_model_parameters=100,
                    max_sequence_length=100,
                    vocabulary_size=100,
                    merge=False,
                    languages=["da"],
                    task="task",
                    results=dict(),
                )
            ],
            True,
        ),
        (
            False,
            True,
            [
                BenchmarkResult(
                    model="model_id@revision",
                    dataset="dataset",
                    generative=False,
                    generative_type=None,
                    few_shot=False,
                    validation_split=False,
                    num_model_parameters=100,
                    max_sequence_length=100,
                    vocabulary_size=100,
                    merge=False,
                    languages=["da"],
                    task="task",
                    results=dict(),
                ),
                BenchmarkResult(
                    model="model_id@revision",
                    dataset="dataset",
                    generative=False,
                    generative_type=None,
                    few_shot=False,
                    validation_split=False,
                    num_model_parameters=100,
                    max_sequence_length=100,
                    vocabulary_size=100,
                    merge=False,
                    languages=["da"],
                    task="task",
                    results=dict(),
                ),
            ],
            True,
        ),
    ],
    ids=[
        "empty benchmark results",
        "model has been benchmarked",
        "model has not been benchmarked",
        "model few-shot has not been benchmarked",
        "model few-shot has been benchmarked",
        "model few-shot has been benchmarked, but not generative",
        "model validation split has not been benchmarked",
        "model validation split has been benchmarked",
        "model has been benchmarked twice",
    ],
)
def test_get_record(
    model_config: ModelConfig,
    dataset_config: DatasetConfig,
    benchmark_config: BenchmarkConfig,
    few_shot: bool,
    evaluate_test_split: bool,
    benchmark_results: list[BenchmarkResult],
    expected: bool,
) -> None:
    """Test whether we can correctly check if a model has been benchmarked."""
    benchmark_config = replace(
        benchmark_config, few_shot=few_shot, evaluate_test_split=evaluate_test_split
    )
    benchmarked = (
        get_record(
            model_config=model_config,
            dataset_config=dataset_config,
            benchmark_config=benchmark_config,
            benchmark_results=benchmark_results,
        )
        is not None
    )
    assert benchmarked == expected


@pytest.mark.parametrize(
    argnames=["verbose", "expected_logging_level"],
    argvalues=[(False, logging.INFO), (True, logging.DEBUG)],
)
def test_adjust_logging_level(verbose: bool, expected_logging_level: int) -> None:
    """Test that the logging level is adjusted correctly."""
    logging_level = adjust_logging_level(verbose=verbose, ignore_testing=True)
    assert logging_level == expected_logging_level


class TestClearCacheFn:
    """Tests related to the `clear_cache_fn` function."""

    def test_clear_non_existing_cache(self) -> None:
        """Test that no errors are thrown when clearing a non-existing cache."""
        clear_model_cache_fn(cache_dir="does-not-exist")
        rmtree(path="does-not-exist", ignore_errors=True)

    def test_clear_existing_cache(self) -> None:
        """Test that a cache can be cleared."""
        cache_dir = Path(".test_euroeval_cache")
        model_cache_dir = cache_dir / "model_cache"
        example_model_dir = model_cache_dir / "example_model"
        dir_to_be_deleted = example_model_dir / "dir_to_be_deleted"

        dir_to_be_deleted.mkdir(parents=True, exist_ok=True)
        assert dir_to_be_deleted.exists()

        clear_model_cache_fn(cache_dir=cache_dir.as_posix())
        assert not dir_to_be_deleted.exists()
        assert example_model_dir.exists()

        rmtree(path=cache_dir, ignore_errors=True)


@pytest.mark.parametrize(
    argnames=["dataset_names", "dataset_configs"],
    argvalues=[
        ([], []),
        (
            ["multi-wiki-qa-da"],
            [
                get_dataset_config(
                    dataset_name="multi-wiki-qa-da",
                    custom_datasets_file=Path("custom_datasets.py"),
                )
            ],
        ),
        (
            ["multi-wiki-qa-da", "dansk"],
            [
                get_dataset_config(
                    dataset_name="multi-wiki-qa-da",
                    custom_datasets_file=Path("custom_datasets.py"),
                ),
                get_dataset_config(
                    dataset_name="dansk",
                    custom_datasets_file=Path("custom_datasets.py"),
                ),
            ],
        ),
    ],
)
def test_prepare_dataset_configs(
    dataset_names: list[str], dataset_configs: list[DatasetConfig]
) -> None:
    """Test that the `prepare_dataset_configs` function works as expected."""
    assert set(
        prepare_dataset_configs(
            dataset_names=dataset_names, custom_datasets_file=Path("custom_datasets.py")
        )
    ) == set(dataset_configs)



# /// script
# requires-python = ">=3.10,<4.0"
# dependencies = [
#     "datasets==3.5.0",
#     "huggingface-hub==0.24.0",
#     "pandas==2.2.0",
#     "requests==2.32.3",
# ]
# ///

"""Create the FoSent sentiment dataset and upload it to the HF Hub."""

import hashlib
import logging
from typing import Literal

import pandas as pd
from datasets import Dataset, DatasetDict, Split, load_dataset
from huggingface_hub import HfApi

from .constants import MAX_NUM_CHARS_IN_DOCUMENT, MIN_NUM_CHARS_IN_DOCUMENT  # noqa

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("create_fosent")


def main() -> None:
    """Create the FoSent sentiment dataset and upload it to the HF Hub."""
    # Define the base download URL
    repo_id = "hafsteinn/faroese_sentiment_analysis"

    # Download the dataset
    dataset = load_dataset(path=repo_id, split="train", token=True)
    assert isinstance(dataset, Dataset)

    # Convert the dataset to a dataframe
    df = dataset.to_pandas()
    assert isinstance(df, pd.DataFrame)

    column_mapping = {
        "News article": "news",
        "Selected Sentence": "sentence",
        "Sentence label - Annotator 1": "sentence_label_1",
        "Sentence label - Annotator 2": "sentence_label_2",
        "News label - Annotator 1": "news_label_1",
        "News label - Annotator 2": "news_label_2",
    }
    df.rename(columns=column_mapping, inplace=True)
    df.drop(
        columns=[col for col in df.columns if col not in column_mapping.values()],
        inplace=True,
    )

    df["news_id"] = df["news"].map(
        lambda x: x if x is None else hashlib.md5(string=x.encode()).hexdigest()
    )
    df["sentence_id"] = df["sentence"].map(
        lambda x: x if x is None else hashlib.md5(string=x.encode()).hexdigest()
    )

    # Create news dataframe
    news_df = df[["news_id", "news", "news_label_1", "news_label_2"]].copy()
    assert isinstance(news_df, pd.DataFrame)
    news_df = (
        news_df.rename(
            columns=dict(
                news_id="id",
                news="text",
                news_label_1="label_1",
                news_label_2="label_2",
            )
        )
        .drop_duplicates(subset="id")
        .reset_index(drop=True)
    )

    # Create sentence dataframe
    sentence_df = df[
        ["sentence_id", "news_id", "sentence", "sentence_label_1", "sentence_label_2"]
    ].copy()
    assert isinstance(sentence_df, pd.DataFrame)
    sentence_df = (
        sentence_df.rename(
            columns=dict(
                sentence_id="id",
                sentence="text",
                sentence_label_1="label_1",
                sentence_label_2="label_2",
            )
        )
        .drop_duplicates(subset="id")
        .reset_index(drop=True)
    )

    # Merge the labels
    news_df["label"] = news_df.apply(
        lambda row: merge_labels(label_1=row.label_1, label_2=row.label_2), axis=1
    )
    sentence_df["label"] = sentence_df.apply(
        lambda row: merge_labels(label_1=row.label_1, label_2=row.label_2), axis=1
    )

    news_df = (
        news_df.drop(columns=["label_1", "label_2"]).dropna().reset_index(drop=True)
    )
    sentence_df = (
        sentence_df.drop(columns=["label_1", "label_2"]).dropna().reset_index(drop=True)
    )

    # Select train/val/test news IDs
    all_news_ids = set(news_df.sample(frac=1, random_state=4242).id)
    assert all_news_ids == set(news_df.id) | set(sentence_df.news_id)
    assert len(all_news_ids) == 170

    # Create validation split
    val_size = 16
    val_news_ids = list(all_news_ids)[:val_size]
    val_df = pd.concat(
        objs=[
            news_df[news_df.id.isin(val_news_ids)].drop(columns=["id"]),
            sentence_df[sentence_df.news_id.isin(val_news_ids)].drop(
                columns=["id", "news_id"]
            ),
        ]
    ).sample(frac=1, random_state=4242)
    assert isinstance(val_df, pd.DataFrame)

    # Create train split
    train_size = 32
    train_news_ids = list(all_news_ids)[val_size : val_size + train_size]
    train_df = pd.concat(
        objs=[
            news_df[news_df.id.isin(train_news_ids)].drop(columns=["id"]),
            sentence_df[sentence_df.news_id.isin(train_news_ids)].drop(
                columns=["id", "news_id"]
            ),
        ]
    ).sample(frac=1, random_state=4242)
    assert isinstance(train_df, pd.DataFrame)

    # Create test split
    test_news_ids = list(all_news_ids)[val_size + train_size :]
    test_df = pd.concat(
        objs=[
            news_df[news_df.id.isin(test_news_ids)].drop(columns=["id"]),
            sentence_df[sentence_df.news_id.isin(test_news_ids)].drop(
                columns=["id", "news_id"]
            ),
        ]
    ).sample(frac=1, random_state=4242)
    assert isinstance(test_df, pd.DataFrame)

    # Reset the index
    train_df = train_df.reset_index(drop=True)
    val_df = val_df.reset_index(drop=True)
    test_df = test_df.reset_index(drop=True)

    # Only work with samples where the document is not very large or small
    # We do it after we have made the splits to ensure that the dataset is minimally
    # affected.
    new_train_df = train_df.copy()
    new_train_df["text_len"] = new_train_df.text.str.len()
    new_train_df = new_train_df.query("text_len >= @MIN_NUM_CHARS_IN_DOCUMENT").query(
        "text_len <= @MAX_NUM_CHARS_IN_DOCUMENT"
    )
    new_val_df = val_df.copy()
    new_val_df["text_len"] = new_val_df.text.str.len()
    new_val_df = new_val_df.query("text_len >= @MIN_NUM_CHARS_IN_DOCUMENT").query(
        "text_len <= @MAX_NUM_CHARS_IN_DOCUMENT"
    )
    new_test_df = test_df.copy()
    new_test_df["text_len"] = new_test_df.text.str.len()
    new_test_df = new_test_df.query("text_len >= @MIN_NUM_CHARS_IN_DOCUMENT").query(
        "text_len <= @MAX_NUM_CHARS_IN_DOCUMENT"
    )

    dataset = DatasetDict(
        {
            "train": Dataset.from_pandas(new_train_df, split=Split.TRAIN),
            "val": Dataset.from_pandas(new_val_df, split=Split.VALIDATION),
            "test": Dataset.from_pandas(new_test_df, split=Split.TEST),
        }
    )

    dataset_id = "EuroEval/fosent"

    # Remove the dataset from Hugging Face Hub if it already exists
    HfApi().delete_repo(dataset_id, repo_type="dataset", missing_ok=True)

    # Push the dataset to the Hugging Face Hub
    dataset.push_to_hub(dataset_id, private=True)


def merge_labels(
    label_1: Literal[-1, 0, 1] | float, label_2: Literal[-1, 0, 1] | float
) -> Literal["negative", "neutral", "positive"] | None:
    """Merge two labels.

    This follows the following rules:
        - If one of the labels is missing, return the other label.
        - If both labels are missing, do not use the label.
        - If the labels are the same, return the label.
        - If the labels are adjacent, return the more extreme label.
        - Otherwise, do not use the label.

    Args:
        label_1:
            The first label.
        label_2:
            The second label.

    Returns:
        The merged label.
    """
    labels: list[Literal["negative", "neutral", "positive"]] = [
        "negative",
        "neutral",
        "positive",
    ]

    if label_1 != label_1 and label_2 == label_2:
        return labels[int(label_2) + 1]
    elif label_1 == label_1 and label_2 != label_2:
        return labels[int(label_1) + 1]
    elif label_1 != label_1 and label_2 != label_2:
        return None

    if label_1 == label_2:
        return labels[int(label_1) + 1]
    elif abs(label_1 - label_2) == 1:
        if max(label_1, label_2) == 1:
            return "positive"
        return "negative"

    return None


if __name__ == "__main__":
    main()



# /// script
# requires-python = ">=3.10,<4.0"
# dependencies = [
#     "datasets==3.5.0",
#     "huggingface-hub==0.24.0",
#     "pandas==2.2.0",
#     "requests==2.32.3",
# ]
# ///

"""Create the LR-Sum summarisation datasets."""

import pandas as pd
from datasets import Dataset, DatasetDict, Split, load_dataset
from huggingface_hub import HfApi

from .constants import MAX_NUM_CHARS_IN_ARTICLE, MIN_NUM_CHARS_IN_ARTICLE

LANGUAGES = {"uk": "ukr", "sr": "srp", "bs": "bos", "sq": "sqi"}


def main() -> None:
    """Create the LR-Sum summarisation mini datasets and upload to HF Hub."""
    dataset_id = "bltlab/lr-sum"

    for language, subset in LANGUAGES.items():
        # Load the full dataset and filter for the subset
        dataset = load_dataset(dataset_id, subset)
        assert isinstance(dataset, DatasetDict)

        dataset = dataset.map(make_columns)
        train_df = dataset["train"].to_pandas()
        val_df = dataset["validation"].to_pandas()
        test_df = dataset["test"].to_pandas()
        assert isinstance(train_df, pd.DataFrame)
        assert isinstance(val_df, pd.DataFrame)
        assert isinstance(test_df, pd.DataFrame)

        train_df = process_df(df=train_df)
        val_df = process_df(df=val_df)
        test_df = process_df(df=test_df)

        # Make splits
        test_size = 2048
        val_size = 256
        train_size = 1024

        # Calculate how many additional samples needed for test split
        additional_test_samples_needed = test_size - len(test_df)
        if additional_test_samples_needed > 0:
            # Take additional samples from training set for test split
            additional_test_samples = train_df.sample(
                n=additional_test_samples_needed, random_state=4242
            )
        else:
            additional_test_samples = pd.DataFrame()

        # Combine original test with additional samples from train
        final_test_df = pd.concat([test_df, additional_test_samples], ignore_index=True)

        # Remove the additional samples from train set
        remaining_train_df = train_df[
            ~train_df.index.isin(additional_test_samples.index)
        ]

        # Sample final splits
        final_test_df = final_test_df.sample(
            n=test_size, random_state=4242
        ).reset_index(drop=True)
        final_val_df = val_df.sample(n=val_size, random_state=4242).reset_index(
            drop=True
        )
        final_train_df = remaining_train_df.sample(
            n=train_size, random_state=4242
        ).reset_index(drop=True)

        assert isinstance(final_train_df, pd.DataFrame)
        assert isinstance(final_val_df, pd.DataFrame)
        assert isinstance(final_test_df, pd.DataFrame)

        # Collect datasets in a dataset dictionary
        mini_dataset = DatasetDict(
            {
                "train": Dataset.from_pandas(final_train_df, split=Split.TRAIN),
                "val": Dataset.from_pandas(final_val_df, split=Split.VALIDATION),
                "test": Dataset.from_pandas(final_test_df, split=Split.TEST),
            }
        )

        # Create dataset ID
        mini_dataset_id = f"EuroEval/lr-sum-{language}-mini"

        # Remove the dataset from Hugging Face Hub if it already exists
        HfApi().delete_repo(mini_dataset_id, repo_type="dataset", missing_ok=True)

        # Push the dataset to the Hugging Face Hub
        mini_dataset.push_to_hub(mini_dataset_id, private=True)


def make_columns(sample: dict) -> dict:
    """Map the dataset to have the text and target_text columns.

    Args:
        sample: A sample from the dataset.

    Returns:
        A sample with the text and target_text columns.
    """
    sample["text"] = f"{sample['title']}\n\n{sample['text']}"
    sample["target_text"] = sample["summary"]
    return sample


def process_df(df: pd.DataFrame) -> pd.DataFrame:
    """Process the dataframe.

    Args:
        df: A dataframe to process.

    Returns:
        Processed dataframe.
    """
    lengths = df.text.str.len()
    lower_bound = MIN_NUM_CHARS_IN_ARTICLE
    upper_bound = MAX_NUM_CHARS_IN_ARTICLE
    df = df.loc[lengths.between(lower_bound, upper_bound)]

    keep_columns = ["text", "target_text"]
    df = df[keep_columns]

    df = df.reset_index(drop=True)
    return df


if __name__ == "__main__":
    main()



"""Tests for the `bias_metrics` module."""

from typing import Callable, Sequence

import numpy as np
import pytest
from datasets import Dataset

from euroeval.metrics import (
    accuracy_ambig_metric,
    bias_adjusted_accuracy_ambig_metric,
    bias_ambig_metric,
)


@pytest.fixture(scope="module")
def make_dataset() -> Callable[[str, Sequence[int], int | None, int], Dataset]:
    """Build small datasets with the columns needed by bias metrics."""

    def _make(
        cond: str, indices: Sequence[int], correct: int | None, n: int
    ) -> Dataset:
        """Create a dataset with consistent context and label indices."""
        stereo, counter, unknown = indices
        records = []
        for _ in range(n):
            records.append(
                {
                    "context_condition": cond,
                    "stereo_idx": stereo,
                    "counter_idx": counter,
                    "unknown_idx": unknown,
                    "correct_idx": correct,
                }
            )
        return Dataset.from_list(records)

    return _make


# --- BiasA tests (ambiguous contexts) ---


def test_biasa_all_unknowns(
    make_dataset: Callable[[str, Sequence[int], int | None, int], Dataset],
) -> None:
    """Return zero bias when only unknown predictions appear."""
    ds = make_dataset("ambig", [1, 2, 0], None, 3)
    preds = [0, 0, 0]
    # references arg is ignored by the metric, so we can pass an empty list
    assert bias_ambig_metric(preds, [], ds, None, None) == pytest.approx(0.0)


def test_biasa_stereo_and_counter(
    make_dataset: Callable[[str, Sequence[int], int | None, int], Dataset],
) -> None:
    """Compute bias from the difference between stereo and counter picks."""
    ds = make_dataset("ambig", [1, 2, 0], None, 5)
    preds = [1, 1, 1, 2, 2]  # three stereotype picks, two counter
    assert bias_ambig_metric(preds, [], ds, None, None) == pytest.approx((3 - 2) / 5)


def test_biasa_ignores_undetectable(
    make_dataset: Callable[[str, Sequence[int], int | None, int], Dataset],
) -> None:
    """Ignore predictions that are not present in label mappings."""
    # one prediction not in label mapping => discarded,
    # but still counts toward n_ambiguous
    ds = make_dataset("ambig", [1, 2, 0], None, 3)
    preds = [1, "X", 2]
    # stereo=1, counter=2 => (1-1)/3 = 0.0
    assert bias_ambig_metric(preds, [], ds, None, None) == pytest.approx(0.0)


# --- AccuracyA tests (ambiguous contexts) ---


def test_accuracy_ambig_all_unknown(
    make_dataset: Callable[[str, Sequence[int], int | None, int], Dataset],
) -> None:
    """Report perfect accuracy when all ambiguous predictions are unknown."""
    ds = make_dataset("ambig", [1, 2, 0], None, 4)
    preds = [0, 0, 0, 0]  # all correctly 'unknown'
    assert accuracy_ambig_metric(preds, [], ds, None, None) == pytest.approx(1.0)


def test_accuracy_ambig_mixed(
    make_dataset: Callable[[str, Sequence[int], int | None, int], Dataset],
) -> None:
    """Compute accuracy with a mix of correct unknown and wrong answers."""
    ds = make_dataset("ambig", [1, 2, 0], None, 4)
    preds = [0, 1, 0, 2]  # two correct 'unknown', two wrong
    assert accuracy_ambig_metric(preds, [], ds, None, None) == pytest.approx(0.5)


def test_accuracy_ambig_accepts_numpy_ints(
    make_dataset: Callable[[str, Sequence[int], int | None, int], Dataset],
) -> None:
    """Accept numpy integer predictions."""
    ds = make_dataset("ambig", [1, 2, 0], None, 3)
    preds = np.array([0, 0, 0], dtype=np.int64)
    assert accuracy_ambig_metric(preds, [], ds, None, None) == pytest.approx(1.0)


# --- Bias-adjusted accuracy metrics ---


def test_bias_adjusted_accuracy_ambig_zero_at_chance(
    make_dataset: Callable[[str, Sequence[int], int | None, int], Dataset],
) -> None:
    """Keep accuracy unchanged when bias is zero at chance."""
    ds = make_dataset("ambig", [1, 2, 0], None, 3)
    preds = [
        0,
        1,
        2,
    ]  # 1 correct unknown, one stereo, one counter => acc = 1/3, bias = 0
    assert bias_adjusted_accuracy_ambig_metric(
        preds, [], ds, None, None
    ) == pytest.approx(1 / 3)


def test_bias_adjusted_accuracy_ambig_above_chance_low_bias(
    make_dataset: Callable[[str, Sequence[int], int | None, int], Dataset],
) -> None:
    """Return accuracy when ambiguous bias stays at zero."""
    ds = make_dataset("ambig", [1, 2, 0], None, 4)
    preds = [0, 0, 1, 2]  # acc = 0.5, bias = 0
    assert bias_adjusted_accuracy_ambig_metric(
        preds, [], ds, None, None
    ) == pytest.approx(0.5)


def test_bias_adjusted_accuracy_ambig_penalizes_bias(
    make_dataset: Callable[[str, Sequence[int], int | None, int], Dataset],
) -> None:
    """Reduce bias-adjusted accuracy when ambiguous bias is high."""
    ds = make_dataset("ambig", [1, 2, 0], None, 4)
    # 2 correct unknown, 2 stereo picks => acc = 0.5, bias = 0.5
    preds = [0, 0, 1, 1]
    # bias_adjusted = max(0, 0.5 - 0.5) = 0.0
    assert bias_adjusted_accuracy_ambig_metric(
        preds, [], ds, None, None
    ) == pytest.approx(0.0)



# /// script
# requires-python = ">=3.10,<4.0"
# dependencies = [
#     "datasets==4.4.2",
#     "huggingface-hub==0.36.0",
#     "pandas==2.2.3",
#     "requests==2.32.3",
#     "evaluate==0.4.6",
#     "bert-score==0.3.13",
# ]
# ///

"""Create the DuidelijkeTaal simplification dataset by filtering the original dataset.

Original data is published at http://hdl.handle.net/10032/tm-a2-y8 and
https://huggingface.co/datasets/instituutnederlandsetaal/DuidelijkeTaal-v1.0.

We use the splitted version of this dataset created during the GPT-NL project
(https://huggingface.co/datasets/GPT-NL/DuidelijkeTaal-v1.0-split).
"""

import pandas as pd
from datasets import ClassLabel, Dataset, DatasetDict, load_dataset
from evaluate import load
from huggingface_hub import HfApi
from requests import HTTPError


def main() -> None:
    """Creates Duidelijke Taal simplification dataset and uploads it to the HF Hub."""
    dataset_id = "GPT-NL/DuidelijkeTaal-v1.0-split"
    dataset_original = load_dataset(dataset_id, token=True)

    # rename columns for easier use in filtering
    column_mapping = {
        "Wordpress ID": "wordpress_id",
        "Document ID": "doc_id",
        "Niet synthetische tekst/zin (A)": "text",
        "Synthetische tekst/zin (B)": "target_text",
        "Paarsgewijze vergelijking": "clarity",
        "Paarsgewijze vergelijking Gem.": "clarity_avg",
        "Accuratesse": "acc",
        "Accuratesse Gem.": "acc_avg",
        "Fluency (A)": "fluency_a",
        "Fluency (A) Gem.": "fluency_a_avg",
        "Fluency (B)": "fluency_b",
        "Fluency (B) Gem.": "fluency_b_avg",
        "Complexiteit (A)": "complexity_a",
        "Complexiteit (A) Gem.": "complexity_a_avg",
        "Complexiteit (B)": "complexity_b",
        "Complexiteit (B) Gem.": "complexity_b_avg",
    }

    dataset_filtered = DatasetDict()
    for split in ["train", "test"]:
        df = dataset_original[split].to_pandas().rename(columns=column_mapping)
        df_filtered = filter_dataset(df)
        dataset_filtered[split] = Dataset.from_pandas(df_filtered).cast_column(
            "complexity_text", ClassLabel(num_classes=4)
        )

    # create 50%/50% train/val split
    # original split is 50%/50% train/test, so total split becomes 25%/25%/50%
    train_val = dataset_filtered["train"].train_test_split(
        test_size=0.5, seed=42, stratify_by_column="complexity_text"
    )

    drop_columns = [
        col
        for col in dataset_filtered["train"].column_names
        if col not in ["text", "target_text"]
    ]

    dataset = DatasetDict(
        {
            "train": train_val["train"].remove_columns(drop_columns),
            "val": train_val["test"].remove_columns(drop_columns),
            "test": dataset_filtered["test"].remove_columns(drop_columns),
        }
    )
    processed_dataset_id = "EuroEval/duidelijke-taal"

    # Remove the dataset from Hugging Face Hub if it already exists
    try:
        api: HfApi = HfApi()
        api.delete_repo(processed_dataset_id, repo_type="dataset")
    except HTTPError:
        pass

    # Push the dataset to the Hugging Face Hub
    dataset.push_to_hub(processed_dataset_id, private=True)


def filter_dataset(
    df: pd.DataFrame,
    min_complexity_difference: int = 10,
    min_n_participants: int = 2,
    min_acc_avg: int = 60,
    min_bertscore: float = 0.76,
    drop_explanations: bool = True,
    verbosity_threshold: int = 2,
) -> pd.DataFrame:
    """Filter the Duidelijke Taal dataset.

    The dataset is filtered based on the complexity difference between the original
    sentence and its synthesized simplification, the number of human annotations,
    and the average accuracy of the simplification.

    Args:
        df (pandas.DataFrame): The unfiltered Duidelijke Taal dataset.
        min_complexity_difference (int, optional): The minimum difference in
            complexity between the original sentence and the simplification to
            create a subset with clearly differentiated examples. Defaults to 10.
        min_n_participants (int, optional): The minimum number of participants from
            the crowd-sourcing experiment who rated the sentence pair, ensuring a
            somewhat general opinion. Defaults to 2.
        min_acc_avg (int, optional): The minimum average accuracy percentage of the
            simplification used to filter out inaccurate simplifications.
            Defaults to 60.
        min_bertscore (float, optional): The minimum BERTScore (F1) similarity
            between the sentence pairs to filter out pairs that vary too much in
            meaning. Defaults to 0.76.
        drop_explanations (bool, optional): Whether to drop sentence pairs where the
            synthetic example contains explanations not present in the original
            sentence (e.g., "Dat betekent dat..."). Defaults to True.
        verbosity_threshold (int, optional): Threshold used to drop pairs where the
            synthetic example contains overly verbose or long explanations that
            deviate from the original text. Simplifications are dropped if they
            contain more than `threshold * number_of_words` of the input text.
            Setting this parameter to 1 disables this filtering. Defaults to 2.

    Returns:
        pandas.DataFrame: The filtered dataset.
    """
    # load default model (bert-base-multilingual-cased), which is sufficient as filter
    bert_score = load("bertscore")
    df["bert"] = bert_score.compute(
        references=df["text"].to_list(),
        predictions=df["target_text"].to_list(),
        lang="nl",
    )["f1"]

    df["complexity_diff"] = df[df["complexity_a_avg"].astype(str).str.len() > 0][
        "complexity_a_avg"
    ].astype(float) - df[df["complexity_b_avg"].astype(str).str.len() > 0][
        "complexity_b_avg"
    ].astype(float)

    # create binned complexity column of input text for stratified dataset split
    df["complexity_text"] = pd.qcut(
        df["complexity_a_avg"],
        q=4,  # 4 bins
        labels=False,
        duplicates="drop",
    )

    df_mask = (
        (
            (~df["complexity_diff"].isna()) & (df["complexity_diff"] > 0)
        )  # only keep the sentences where original is harder
        & (df["complexity_diff"].abs() > min_complexity_difference)
        & (
            df["complexity_a"].apply(
                lambda x: (len(x) >= min_n_participants if x is not None else False)
            )
        )
        & (
            df["complexity_b"].apply(
                lambda x: (len(x) >= min_n_participants if x is not None else False)
            )
        )
        & (df["acc_avg"].apply(lambda x: x != "" and float(x) >= min_acc_avg))
        & (df["bert"] >= min_bertscore)
    )

    # 'betekent' = 'means' in Dutch
    if drop_explanations:
        df_mask &= ~df["target_text"].str.contains("betekent")

    df_filtered = df[df_mask]

    if verbosity_threshold > 1:
        df_filtered = df_filtered[
            verbosity_threshold * df_filtered["text"].str.split().str.len()
            >= df_filtered["target_text"].str.split().str.len()
        ]

    df_filtered_dedup = df_filtered.drop_duplicates(subset="text")
    print(f"Dropped {len(df_filtered) - len(df_filtered_dedup)} duplicate examples.")
    print(
        f"Filtered dataset split contains {len(df_filtered_dedup)} simplification "
        f"pairs."
    )
    return df_filtered_dedup


if __name__ == "__main__":
    main()



# /// script
# requires-python = ">=3.10,<4.0"
# dependencies = [
#     "datasets==3.5.0",
#     "huggingface-hub==0.24.0",
#     "pandas==2.2.0",
#     "requests==2.32.3",
#     "scikit-learn<1.6.0",
# ]
# ///

"""Create the Global-MMLU knowledge datasets and upload them to the HF Hub."""

from collections import Counter

import pandas as pd
from datasets import Dataset, DatasetDict, Split, load_dataset
from huggingface_hub import HfApi
from sklearn.model_selection import train_test_split

from .constants import (
    CHOICES_MAPPING,
    MAX_NUM_CHARS_IN_INSTRUCTION,
    MAX_NUM_CHARS_IN_OPTION,
    MAX_REPETITIONS,
    MIN_NUM_CHARS_IN_INSTRUCTION,
    MIN_NUM_CHARS_IN_OPTION,
)


def main() -> None:
    """Create the Global-MMLU knowledge datasets."""
    # Define the dataset ID
    repo_id = "CohereLabs/Global-MMLU"

    languages = ["uk", "el", "ro", "sq"]

    for language in languages:
        try:
            # Load the dataset
            dataset = load_dataset(path=repo_id, name=language)
            lite = False
        except ValueError:
            # load the lite dataset
            repo_id = "CohereLabs/Global-MMLU-lite"
            dataset = load_dataset(path=repo_id, name=language)
            lite = True
        assert isinstance(dataset, DatasetDict)

        # Convert the dataset to dataframes
        val_df = dataset["dev"].to_pandas()
        test_df = dataset["test"].to_pandas()
        assert isinstance(val_df, pd.DataFrame)
        assert isinstance(test_df, pd.DataFrame)

        # Process the dataframes
        val_df = process_split(df=val_df, language=language)
        test_df = process_split(df=test_df, language=language)

        if lite:
            train_df, val_df, test_df = make_mmlu_lite_splits(
                val_df=val_df, test_df=test_df
            )
        else:
            train_df, test_df, val_df = make_mmlu_splits(val_df=val_df, test_df=test_df)

        # Collect datasets in a dataset dictionary
        dataset = DatasetDict(
            {
                "train": Dataset.from_pandas(train_df, split=Split.TRAIN),
                "val": Dataset.from_pandas(val_df, split=Split.VALIDATION),
                "test": Dataset.from_pandas(test_df, split=Split.TEST),
            }
        )

        # Create dataset ID
        lite_suffix = "-lite" if lite else ""
        mini_suffix = "-mini" if not lite else ""
        dataset_id = f"EuroEval/global-mmlu{lite_suffix}-{language}{mini_suffix}"
        # Remove the dataset from Hugging Face Hub if it already exists
        HfApi().delete_repo(dataset_id, repo_type="dataset", missing_ok=True)

        # Push the dataset to the Hugging Face Hub
        dataset.push_to_hub(dataset_id, private=True)


def make_mmlu_lite_splits(
    val_df: pd.DataFrame, test_df: pd.DataFrame
) -> tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """Make splits for a lite dataset.

    Args:
        val_df: The validation dataframe.
        test_df: The test dataframe.

    Returns:
        The final training, validation, and test dataframes.
    """
    val_size = 64
    train_size = 128
    train_df_final, test_df = train_test_split(
        test_df, train_size=train_size, random_state=4242, stratify=test_df.category
    )

    val_df_final = val_df.sample(n=val_size, random_state=4242, replace=False)

    # Use remaining val samples in the test split
    remaining_val_df = val_df[~val_df.index.isin(val_df_final.index)]
    test_df_final = pd.concat([test_df, remaining_val_df], ignore_index=True)

    # Reset the index
    train_df_final = train_df_final.reset_index(drop=True)
    val_df_final = val_df_final.reset_index(drop=True)
    test_df_final = test_df_final.reset_index(drop=True)

    return train_df_final, val_df_final, test_df_final


def make_mmlu_splits(
    val_df: pd.DataFrame, test_df: pd.DataFrame
) -> tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """Make splits for a full dataset.

    Args:
        val_df: The validation dataframe.
        test_df: The test dataframe.

    Returns:
        The final training, validation, and test dataframes.
    """
    val_size = 256
    val_df_final = val_df.sample(n=val_size, random_state=4242, replace=False)

    train_size = 1024
    test_size = 2048

    train_df_final, test_df = train_test_split(
        test_df, train_size=train_size, random_state=4242, stratify=test_df.category
    )

    test_df_final = test_df.sample(n=test_size, random_state=4242, replace=False)

    # Reset the index
    train_df_final = train_df_final.reset_index(drop=True)
    val_df_final = val_df_final.reset_index(drop=True)
    test_df_final = test_df_final.reset_index(drop=True)
    return train_df_final, test_df_final, val_df_final


def process_split(df: pd.DataFrame, language: str) -> pd.DataFrame:
    """Process a split of the dataset.

    Args:
        df: The input DataFrame.

    Returns:
        The processed DataFrame.
    """
    df["label"] = df["answer"].str.lower()
    df["category"] = df["subject_category"]
    df = filter_by_length(df=df)
    df = filter_repetitive(df=df)
    df = add_text_column(df=df, language=language)
    df = df.loc[:, ["text", "label", "category"]]
    df = df.drop_duplicates(inplace=False)
    df = df.reset_index(drop=True)
    return df


def filter_by_length(df: pd.DataFrame) -> pd.DataFrame:
    """Remove samples with overly short or long texts.

    Args:
        df: The input DataFrame.

    Returns:
        The filtered DataFrame with only rows of acceptable length.
    """
    return df.loc[
        (df.question.str.len() >= MIN_NUM_CHARS_IN_INSTRUCTION)
        & (df.question.str.len() <= MAX_NUM_CHARS_IN_INSTRUCTION)
        & (df.option_a.str.len() >= MIN_NUM_CHARS_IN_OPTION)
        & (df.option_a.str.len() <= MAX_NUM_CHARS_IN_OPTION)
        & (df.option_b.str.len() >= MIN_NUM_CHARS_IN_OPTION)
        & (df.option_b.str.len() <= MAX_NUM_CHARS_IN_OPTION)
        & (df.option_c.str.len() >= MIN_NUM_CHARS_IN_OPTION)
        & (df.option_c.str.len() <= MAX_NUM_CHARS_IN_OPTION)
        & (df.option_d.str.len() >= MIN_NUM_CHARS_IN_OPTION)
        & (df.option_d.str.len() <= MAX_NUM_CHARS_IN_OPTION)
    ]


def is_repetitive(text: str) -> bool:
    """Check if a text is repetitive.

    Args:
        text: The input text string.

    Returns:
        True if repetitive, False otherwise.
    """
    max_repetitions = max(Counter(text.split()).values())
    return max_repetitions > MAX_REPETITIONS


def filter_repetitive(df: pd.DataFrame) -> pd.DataFrame:
    """Remove overly repetitive samples.

    Args:
        df: The input DataFrame.

    Returns:
        The filtered DataFrame without overly repetitive texts.
    """
    return df.loc[
        ~df.question.apply(is_repetitive)
        & ~df.option_a.apply(is_repetitive)
        & ~df.option_b.apply(is_repetitive)
        & ~df.option_c.apply(is_repetitive)
        & ~df.option_d.apply(is_repetitive)
    ]


def add_text_column(df: pd.DataFrame, language: str) -> pd.DataFrame:
    """Make a `text` column with all the options.

    Args:
        df: The input DataFrame.

    Returns:
        The DataFrame with the added `text` column.
    """
    df["text"] = [
        row.question.replace("\n", " ").strip() + "\n"
        f"{CHOICES_MAPPING[language]}:\n"
        "a. " + row.option_a.replace("\n", " ").strip() + "\n"
        "b. " + row.option_b.replace("\n", " ").strip() + "\n"
        "c. " + row.option_c.replace("\n", " ").strip() + "\n"
        "d. " + row.option_d.replace("\n", " ").strip()
        for _, row in df.iterrows()
    ]
    return df


if __name__ == "__main__":
    main()



"""Templates for the token classification task."""

import typing as t

from ..data_models import PromptConfig
from ..languages import (
    BULGARIAN,
    CZECH,
    DANISH,
    DUTCH,
    ENGLISH,
    ESTONIAN,
    FAROESE,
    FINNISH,
    FRENCH,
    GERMAN,
    GREEK,
    ICELANDIC,
    ITALIAN,
    LATVIAN,
    LITHUANIAN,
    NORWEGIAN,
    NORWEGIAN_BOKMÅL,
    NORWEGIAN_NYNORSK,
    POLISH,
    PORTUGUESE,
    SLOVAK,
    SPANISH,
    SWEDISH,
    UKRAINIAN,
)

if t.TYPE_CHECKING:
    from ..languages import Language


TOKEN_CLASSIFICATION_TEMPLATES: dict["Language", PromptConfig] = {
    ENGLISH: PromptConfig(
        default_prompt_prefix="Below are texts and JSON dictionaries with the "
        "categories that occur in the given text.",
        default_prompt_template="Text: {text}\nCategories: {label}",
        default_instruction_prompt="Text: {text}\n\nIdentify the categories in "
        "the text. You should output this as a JSON dictionary with the keys being "
        "{labels_str}. The values should be lists of the spans of that category, "
        "exactly as they appear in the text.",
        default_prompt_label_mapping="auto",
    ),
    BULGARIAN: PromptConfig(
        default_prompt_prefix="По-долу са текстове и JSON речници с категориите, "
        "които се срещат в дадения текст.",
        default_prompt_template="Текст: {text}\nКатегории: {label}",
        default_instruction_prompt="Текст: {text}\n\nИдентифицирайте категориите "
        "в текста. Трябва да изведете това като JSON речник, като ключовете са "
        "{labels_str}. Стойностите трябва да бъдат списъци с откъсите от тази "
        "категория, точно както се появяват в текста.",
        default_prompt_label_mapping="auto",
    ),
    CZECH: PromptConfig(
        default_prompt_prefix="Níže jsou texty a JSON slovníky s kategoriemi, "
        "které se v daném textu vyskytují.",
        default_prompt_template="Text: {text}\nKategorie: {label}",
        default_instruction_prompt="Text: {text}\n\nIdentifikujte kategorie v "
        "textu. Měli byste to vypsat jako JSON slovník, kde klíče jsou "
        "{labels_str}. Hodnoty by měly být seznamy úseků dané kategorie, přesně "
        "tak, jak se objevují v textu.",
        default_prompt_label_mapping="auto",
    ),
    DANISH: PromptConfig(
        default_prompt_prefix="Nedenfor er tekster og JSON-ordbøger med de "
        "kategorier, der forekommer i den givne tekst.",
        default_prompt_template="Tekst: {text}\nKategorier: {label}",
        default_instruction_prompt="Tekst: {text}\n\nIdentificer kategorierne i "
        "teksten. Du skal udskrive dette som en JSON-ordbog, hvor nøglerne er "
        "{labels_str}. Værdierne skal være lister over uddrag af den kategori, "
        "præcis som de optræder i teksten.",
        default_prompt_label_mapping="auto",
    ),
    GERMAN: PromptConfig(
        default_prompt_prefix="Unten sind Texte und JSON-Wörterbücher mit den "
        "Kategorien, die im jeweiligen Text vorkommen.",
        default_prompt_template="Text: {text}\nKategorien: {label}",
        default_instruction_prompt="Text: {text}\n\nIdentifiziere die Kategorien "
        "im Text. Du solltest dies als ein JSON-Wörterbuch ausgeben, wobei die "
        "Schlüssel {labels_str} sind. Die Werte sollten Listen der Ausschnitte "
        "dieser Kategorie sein, genau so, wie sie im Text erscheinen.",
        default_prompt_label_mapping="auto",
    ),
    GREEK: PromptConfig(
        default_prompt_prefix="Παρακάτω είναι κείμενα και JSON λεξικά με τις "
        "κατηγορίες που εμφανίζονται στο δεδομένο κείμενο.",
        default_prompt_template="Κείμενο: {text}\nΚατηγορίες: {label}",
        default_instruction_prompt="Κείμενο: {text}\n\nΑναγνωρίστε τις "
        "κατηγορίες στο κείμενο. Θα πρέπει να το εκτυπώσετε ως ένα JSON "
        "λεξικό με τα κλειδιά να είναι {labels_str}. Οι τιμές θα πρέπει να "
        "είναι λίστες με τα αποσπάσματα αυτής της κατηγορίας, ακριβώς όπως "
        "εμφανίζονται στο κείμενο.",
        default_prompt_label_mapping="auto",
    ),
    SPANISH: PromptConfig(
        default_prompt_prefix="A continuación se presentan textos y diccionarios "
        "JSON con las categorías que aparecen en el texto dado.",
        default_prompt_template="Texto: {text}\nCategorías: {label}",
        default_instruction_prompt="Texto: {text}\n\nIdentifica las categorías "
        "en el texto. Debes imprimir esto como un diccionario JSON con las "
        "claves siendo {labels_str}. Los valores deben ser listas de los "
        "fragmentos de esa categoría, tal como aparecen en el texto.",
        default_prompt_label_mapping="auto",
    ),
    ESTONIAN: PromptConfig(
        default_prompt_prefix="Allpool on tekstid ja JSON-sõnastikud kategooriatega, "
        "mis esinevad antud tekstis.",
        default_prompt_template="Tekst: {text}\nKategooriad: {label}",
        default_instruction_prompt="Tekst: {text}\n\nTuvastage tekstis "
        "kategooriad. Te peate selle väljatrükkima JSON-sõnastikuna, kus "
        "võtmed on {labels_str}. Väärtused peaksid olema selle kategooria "
        "lõikude loendid, täpselt nii, nagu need tekstis esinevad.",
        default_prompt_label_mapping="auto",
    ),
    FINNISH: PromptConfig(
        default_prompt_prefix="Alla on tekstejä ja JSON-sanakirjoja, joissa on "
        "kategorioita, jotka esiintyvät annetussa tekstissä.",
        default_prompt_template="Teksti: {text}\nKategoriat: {label}",
        default_instruction_prompt="Teksti: {text}\n\nTunnista tekstin "
        "kategoriat. Sinun tulee tulostaa tämä JSON-sanakirjana, jossa "
        "avaimet ovat {labels_str}. Arvojen tulee olla kyseisen kategorian "
        "pätkien listoja, täsmälleen niin kuin ne esiintyvät tekstissä.",
        default_prompt_label_mapping="auto",
    ),
    FAROESE: PromptConfig(
        default_prompt_prefix="Niðanfyri eru tekstir og JSON orðabøkur við "
        "bólkum, ið eru í givna tekstinum.",
        default_prompt_template="Tekstur: {text}\nBólkar: {label}",
        default_instruction_prompt="Tekstur: {text}\n\nKenn aftur bólkarnar "
        "í tekstinum. Tú skalt prenta hetta sum ein JSON orðabók, har "
        "lyklarnir eru {labels_str}. Virðini skulu vera listar yvir "
        "brotini av tí bólkinum, beint sum tey síggjast í tekstinum.",
        default_prompt_label_mapping="auto",
    ),
    FRENCH: PromptConfig(
        default_prompt_prefix="Ci-dessous se trouvent des textes et des dictionnaires "
        "JSON avec les catégories qui apparaissent dans le texte donné.",
        default_prompt_template="Texte : {text}\nCatégories : {label}",
        default_instruction_prompt="Texte : {text}\n\nIdentifiez les catégories "
        "dans le texte. Vous devez l'imprimer sous la forme d'un dictionnaire JSON "
        "avec pour clés {labels_str}. Les valeurs doivent être des listes des "
        "extraits de cette catégorie, exactement comme ils apparaissent dans le texte.",
        default_prompt_label_mapping="auto",
    ),
    ICELANDIC: PromptConfig(
        default_prompt_prefix="Hér fyrir neðan eru textar og JSON orðabækur með "
        "flokkum sem koma fyrir í tilteknum texta.",
        default_prompt_template="Texti: {text}\nFlokkar: {label}",
        default_instruction_prompt="Texti: {text}\n\nFinndu flokkana í "
        "textanum. Þú átt að prenta þetta sem JSON orðabók þar sem lyklar "
        "eru {labels_str}. Gildin eiga að vera listar yfir brot af þeim "
        "flokki, nákvæmlega eins og þau koma fyrir í textanum.",
        default_prompt_label_mapping="auto",
    ),
    ITALIAN: PromptConfig(
        default_prompt_prefix="Di seguito sono riportati testi e dizionari JSON "
        "con le categorie che compaiono nel testo dato.",
        default_prompt_template="Testo: {text}\nCategorie: {label}",
        default_instruction_prompt="Testo: {text}\n\nIdentifica le categorie "
        "nel testo. Devi stampare questo come un dizionario JSON con le chiavi "
        "che sono {labels_str}. I valori devono essere liste dei brani di "
        "quella categoria, esattamente come appaiono nel testo.",
        default_prompt_label_mapping="auto",
    ),
    LITHUANIAN: PromptConfig(
        default_prompt_prefix="Žemiau pateikti tekstai ir JSON žodynai su "
        "kategorijomis, kurios pasitaiko nurodytame tekste.",
        default_prompt_template="Tekstas: {text}\nKategorijos: {label}",
        default_instruction_prompt="Tekstas: {text}\n\nNustatykite kategorijas "
        "tekste. Turite tai atspausdinti kaip JSON žodyną, kur raktai yra "
        "{labels_str}. Reikšmės turėtų būti tos kategorijos ištraukų sąrašai, "
        "tiksliai taip, kaip jos pateikiamos tekste.",
        default_prompt_label_mapping="auto",
    ),
    LATVIAN: PromptConfig(
        default_prompt_prefix="Zemāk ir teksti un JSON vārdnīcas ar kategorijām, "
        "kas parādās dotajā tekstā.",
        default_prompt_template="Teksts: {text}\nKategorijas: {label}",
        default_instruction_prompt="Teksts: {text}\n\nIdentificējiet "
        "kategorijas tekstā. Jums tas jāizdrukā kā JSON vārdnīca, kur "
        "atslēgas ir {labels_str}. Vērtībām jābūt šo kategoriju "
        "izvilkumu sarakstiem, tieši tā, kā tās parādās tekstā.",
        default_prompt_label_mapping="auto",
    ),
    NORWEGIAN_BOKMÅL: PromptConfig(
        default_prompt_prefix="Nedenfor er tekster og JSON-ordbøker med de "
        "kategoriene som forekommer i den gitte teksten.",
        default_prompt_template="Tekst: {text}\nKategorier: {label}",
        default_instruction_prompt="Tekst: {text}\n\nIdentifiser kategoriene "
        "i teksten. Du skal skrive dette ut som en JSON-ordbok med nøklene "
        "som er {labels_str}. Verdiene skal være lister over utdragene av "
        "den kategorien, akkurat som de vises i teksten.",
        default_prompt_label_mapping="auto",
    ),
    DUTCH: PromptConfig(
        default_prompt_prefix="Hieronder volgen teksten en JSON-woordenboeken "
        "met de categorieën die in de gegeven tekst voorkomen.",
        default_prompt_template="Tekst: {text}\nCategorieën: {label}",
        default_instruction_prompt="Tekst: {text}\n\nIdentificeer de "
        "categorieën in de tekst. Je moet dit afdrukken als een JSON-woordenboek "
        "met de sleutels zijnde {labels_str}. De waarden moeten lijsten zijn van "
        "de fragmenten van die categorie, precies zoals ze in de tekst voorkomen.",
        default_prompt_label_mapping="auto",
    ),
    NORWEGIAN_NYNORSK: PromptConfig(
        default_prompt_prefix="Nedanfor er tekstar og JSON-ordbøker med dei "
        "kategoriane som førekjem i den gitte teksten.",
        default_prompt_template="Tekst: {text}\nKategoriar: {label}",
        default_instruction_prompt="Tekst: {text}\n\nIdentifiser kategoriane "
        "i teksten. Du skal skrive dette ut som ein JSON-ordbok med nøklane "
        "som er {labels_str}. Verdiane skal vere lister over utdraga av den "
        "kategorien, akkurat som dei viser seg i teksten.",
        default_prompt_label_mapping="auto",
    ),
    NORWEGIAN: PromptConfig(
        default_prompt_prefix="Nedenfor er tekster og JSON-ordbøker med de "
        "kategoriene som forekommer i den gitte teksten.",
        default_prompt_template="Tekst: {text}\nKategorier: {label}",
        default_instruction_prompt="Tekst: {text}\n\nIdentifiser kategoriene "
        "i teksten. Du skal skrive dette ut som en JSON-ordbok med nøklene "
        "som er {labels_str}. Verdiene skal være lister over utdragene av "
        "den kategorien, akkurat som de vises i teksten.",
        default_prompt_label_mapping="auto",
    ),
    POLISH: PromptConfig(
        default_prompt_prefix="Poniżej znajdują się teksty i słowniki JSON z "
        "kategoriami występującymi w danym tekście.",
        default_prompt_template="Tekst: {text}\nKategorie: {label}",
        default_instruction_prompt="Tekst: {text}\n\nZidentyfikuj kategorie "
        "w tekście. Należy to wydrukować jako słownik JSON, w którym kluczami "
        "są {labels_str}. Wartości powinny być listami fragmentów danej "
        "kategorii, dokładnie tak, jak pojawiają się w tekście.",
        default_prompt_label_mapping="auto",
    ),
    PORTUGUESE: PromptConfig(
        default_prompt_prefix="A seguir estão textos e dicionários JSON com as "
        "categorias que aparecem no texto dado.",
        default_prompt_template="Texto: {text}\nCategorias: {label}",
        default_instruction_prompt="Texto: {text}\n\nIdentifique as categorias "
        "no texto. Você deve imprimir isso como um dicionário JSON com as "
        "chaves sendo {labels_str}. Os valores devem ser listas dos trechos "
        "dessa categoria, exatamente como aparecem no texto.",
        default_prompt_label_mapping="auto",
    ),
    SLOVAK: PromptConfig(
        default_prompt_prefix="Nižšie sú texty a JSON slovníky s kategóriami, "
        "ktoré sa v danom texte vyskytujú.",
        default_prompt_template="Text: {text}\nKategórie: {label}",
        default_instruction_prompt="Text: {text}\n\nIdentifikujte kategórie v "
        "texte. Mali by ste to vypísať ako JSON slovník, kde kľúče sú "
        "{labels_str}. Hodnoty by mali byť zoznamy úsekov danej kategórie, "
        "presne tak, ako sa objavujú v texte.",
        default_prompt_label_mapping="auto",
    ),
    SWEDISH: PromptConfig(
        default_prompt_prefix="Nedan är texter och JSON-ordböcker med de "
        "kategorier som förekommer i den givna texten.",
        default_prompt_template="Text: {text}\nKategorier: {label}",
        default_instruction_prompt="Text: {text}\n\nIdentifiera kategorierna "
        "i texten. Du ska skriva ut detta som en JSON-ordbok med nycklarna "
        "som är {labels_str}. Värdena ska vara listor över utdragen av den "
        "kategorin, precis som de visas i texten.",
        default_prompt_label_mapping="auto",
    ),
    UKRAINIAN: PromptConfig(
        default_prompt_prefix="Нижче наведено тексти та JSON-словники з "
        "категоріями, які зустрічаються в наведеному тексті.",
        default_prompt_template="Текст: {text}\nКатегорії: {label}",
        default_instruction_prompt="Текст: {text}\n\nВизначте категорії в "
        "тексті. Ви повинні надрукувати це як JSON-словник, де ключі - це "
        "{labels_str}. Значення повинні бути списками уривків цієї категорії, "
        "саме так, як вони з'являються в тексті.",
        default_prompt_label_mapping="auto",
    ),
}



# /// script
# requires-python = ">=3.10,<4.0"
# dependencies = [
#     "click==8.1.3",
#     "datasets==3.5.0",
#     "huggingface-hub==0.24.0",
#     "pandas==2.2.0",
#     "requests==2.32.3",
# ]
# ///

"""Create the FQuAD-mini NER dataset and upload it to the HF Hub."""

from collections import defaultdict
from pathlib import Path

import click
import pandas as pd
from datasets import Dataset, DatasetDict, Split
from huggingface_hub import HfApi

from .constants import (
    MAX_NUM_CHARS_IN_CONTEXT,
    MAX_NUM_CHARS_IN_QUESTION,
    MIN_NUM_CHARS_IN_CONTEXT,
    MIN_NUM_CHARS_IN_QUESTION,
)


@click.command(
    help="Create the FQuAD-mini NER dataset and upload it to the HF Hub. Note that you "
    "need to request the raw data from https://fquad.illuin.tech/#download before you "
    "can run this script."
)
@click.argument("data_dir", type=str)
def main(data_dir: str) -> None:
    """Create the FQuAD-mini NER dataset and uploads it to the HF Hub."""
    train_path = Path(data_dir) / "train.json"
    train_df = pd.read_json(train_path)

    valtest_path = Path(data_dir) / "valid.json"
    valtest_df = pd.read_json(valtest_path)

    def set_up_dataframe(df: pd.DataFrame) -> pd.DataFrame:
        """Set up the dataframe."""
        data_dict: dict[str, list] = defaultdict(list)
        for article_dict in df.data:
            for paragraph_dict in article_dict["paragraphs"]:
                context = paragraph_dict["context"]
                for qa_dict in paragraph_dict["qas"]:
                    id_ = qa_dict["id"]
                    question = qa_dict["question"]
                    answer_dict = qa_dict["answers"]
                    data_dict["id"].append(id_)
                    data_dict["context"].append(context)
                    data_dict["question"].append(question)
                    data_dict["answers"].append(
                        dict(
                            text=[answer["text"] for answer in answer_dict],
                            answer_start=[
                                answer["answer_start"] for answer in answer_dict
                            ],
                        )
                    )
        return pd.DataFrame(data_dict)

    train_df = set_up_dataframe(df=train_df)
    valtest_df = set_up_dataframe(df=valtest_df)

    # Remove duplicates
    train_df = train_df.drop_duplicates(subset="id")
    valtest_df = valtest_df.drop_duplicates(subset="id")

    # Only work with samples where the context is not very large or small
    train_lengths = train_df.context.str.len()
    valtest_lengths = valtest_df.context.str.len()
    train_df = train_df.loc[
        train_lengths.between(MIN_NUM_CHARS_IN_CONTEXT, MAX_NUM_CHARS_IN_CONTEXT)
    ]
    valtest_df = valtest_df.loc[
        valtest_lengths.between(MIN_NUM_CHARS_IN_CONTEXT, MAX_NUM_CHARS_IN_CONTEXT)
    ]

    # Only work with samples where the question is not very large or small
    train_question_lengths = train_df.question.str.len()
    valtest_question_lengths = valtest_df.question.str.len()
    train_df = train_df[
        train_question_lengths.between(
            MIN_NUM_CHARS_IN_QUESTION, MAX_NUM_CHARS_IN_QUESTION
        )
    ]
    valtest_df = valtest_df[
        valtest_question_lengths.between(
            MIN_NUM_CHARS_IN_QUESTION, MAX_NUM_CHARS_IN_QUESTION
        )
    ]

    # Extract information on which examples contain an answer
    def has_answer(example: dict) -> bool:
        return len(example["text"]) > 0 and example["text"][0] != ""

    train_has_answer: pd.Series = train_df.answers.map(has_answer)
    valtest_has_answer: pd.Series = valtest_df.answers.map(has_answer)

    # Only work with the questions having answers in the context
    train_df_with_answer: pd.DataFrame = train_df.loc[train_has_answer]
    valtest_df_with_answer: pd.DataFrame = valtest_df.loc[valtest_has_answer]

    # Create validation split
    val_size = 256
    val_df = valtest_df_with_answer.sample(n=val_size, random_state=4242)

    # Create test split
    test_size = 2048
    valtest_df_with_answer_filtered: pd.DataFrame = valtest_df_with_answer.loc[
        ~valtest_df_with_answer.index.isin(val_df.index)
    ]
    test_df = valtest_df_with_answer_filtered.sample(n=test_size, random_state=4242)

    # Create train split
    train_size = 1024
    train_df = train_df_with_answer.sample(n=train_size, random_state=4242)

    val_df = val_df.reset_index(drop=True)
    test_df = test_df.reset_index(drop=True)
    train_df = train_df.reset_index(drop=True)

    # Collect datasets in a dataset dictionary
    dataset = DatasetDict(
        {
            "train": Dataset.from_pandas(train_df, split=Split.TRAIN),
            "val": Dataset.from_pandas(val_df, split=Split.VALIDATION),
            "test": Dataset.from_pandas(test_df, split=Split.TEST),
        }
    )

    def add_answer_start(example: dict) -> dict:
        possible_answers: list[str] = example["answers"]["text"]
        if possible_answers:
            answers: list[str] = list()
            answer_starts: list[int] = list()
            for answer in possible_answers:
                answer_start = example["context"].find(answer)
                if answer_start >= 0:
                    answers.append(answer)
                    answer_starts.append(answer_start)
            example["answers"]["text"] = answers
            example["answers"]["answer_start"] = answer_starts
        else:
            example["answers"]["answer_start"] = []
        return example

    dataset = dataset.map(function=add_answer_start)

    # Create dataset ID
    mini_dataset_id = "EuroEval/fquad-mini"

    # Remove the dataset from Hugging Face Hub if it already exists
    HfApi().delete_repo(mini_dataset_id, repo_type="dataset", missing_ok=True)

    # Push the dataset to the Hugging Face Hub
    dataset.push_to_hub(mini_dataset_id, private=True)


if __name__ == "__main__":
    main()



"""All Spanish dataset configurations used in EuroEval."""

from ..data_models import DatasetConfig
from ..languages import SPANISH
from ..tasks import COMMON_SENSE, EUROPEAN_VALUES, KNOW, LA, MCRC, NER, RC, SENT, SUMM

### Official datasets ###

SENTIMENT_HEADLINES_CONFIG = DatasetConfig(
    name="sentiment-headlines-es",
    pretty_name="Sentiment Headlines ES",
    source="EuroEval/sentiment-headlines-es",
    task=SENT,
    languages=[SPANISH],
)

SCALA_ES_CONFIG = DatasetConfig(
    name="scala-es",
    pretty_name="ScaLA-es",
    source="EuroEval/scala-es",
    task=LA,
    languages=[SPANISH],
)

CONLL_ES_CONFIG = DatasetConfig(
    name="conll-es",
    pretty_name="CoNLL-es",
    source="EuroEval/conll-es-mini",
    task=NER,
    languages=[SPANISH],
)

MLQA_ES_CONFIG = DatasetConfig(
    name="mlqa-es",
    pretty_name="MLQA-es",
    source="EuroEval/mlqa-es",
    task=RC,
    languages=[SPANISH],
)

MLSUM_ES_CONFIG = DatasetConfig(
    name="mlsum-es",
    pretty_name="MLSUM-es",
    source="EuroEval/mlsum-es-mini",
    task=SUMM,
    languages=[SPANISH],
)

MMLU_ES_CONFIG = DatasetConfig(
    name="mmlu-es",
    pretty_name="MMLU-es",
    source="EuroEval/mmlu-es-mini",
    task=KNOW,
    languages=[SPANISH],
)

HELLASWAG_ES_CONFIG = DatasetConfig(
    name="hellaswag-es",
    pretty_name="HellaSwag-es",
    source="EuroEval/hellaswag-es-mini",
    task=COMMON_SENSE,
    languages=[SPANISH],
)

VALEU_ES_CONFIG = DatasetConfig(
    name="valeu-es",
    pretty_name="VaLEU-es",
    source="EuroEval/european-values-es",
    task=EUROPEAN_VALUES,
    languages=[SPANISH],
    splits=["test"],
    bootstrap_samples=False,
    _instruction_prompt="{text}",
)


### Unofficial datasets ###

XQUAD_ES_CONFIG = DatasetConfig(
    name="xquad-es",
    pretty_name="XQuAD-es",
    source="EuroEval/xquad-es",
    task=RC,
    languages=[SPANISH],
    unofficial=True,
)

BELEBELE_ES_CONFIG = DatasetConfig(
    name="belebele-es",
    pretty_name="Belebele-es",
    source="EuroEval/belebele-es-mini",
    task=MCRC,
    languages=[SPANISH],
    unofficial=True,
)

MULTI_WIKI_QA_ES_CONFIG = DatasetConfig(
    name="multi-wiki-qa-es",
    pretty_name="MultiWikiQA-es",
    source="EuroEval/multi-wiki-qa-es-mini",
    task=RC,
    languages=[SPANISH],
    unofficial=True,
)

DACSA_ES_CONFIG = DatasetConfig(
    name="dacsa-es",
    pretty_name="DACSA-es",
    source="EuroEval/dacsa-es-mini",
    task=SUMM,
    languages=[SPANISH],
    unofficial=True,
)

GOLDENSWAG_ES_CONFIG = DatasetConfig(
    name="goldenswag-es",
    pretty_name="GoldenSwag-es",
    source="EuroEval/goldenswag-es-mini",
    task=COMMON_SENSE,
    languages=[SPANISH],
    unofficial=True,
)

WINOGRANDE_ES_CONFIG = DatasetConfig(
    name="winogrande-es",
    pretty_name="Winogrande-es",
    source="EuroEval/winogrande-es",
    task=COMMON_SENSE,
    languages=[SPANISH],
    _labels=["a", "b"],
    unofficial=True,
)



"""Bias and accuracy metrics for the MBBQ dataset."""

import collections.abc as c
import numbers
import typing as t

from .base import Metric

if t.TYPE_CHECKING:
    from datasets.arrow_dataset import Dataset

    from ..data_models import BenchmarkConfig, DatasetConfig

BiasType = t.Literal["bias_ambig", "accuracy_ambig", "bias_adjusted_accuracy_ambig"]
VALID_BIAS_TYPES: tuple[BiasType, ...] = t.get_args(BiasType)

CHOICE_TO_INDEX: dict[str, int] = {"a": 0, "b": 1, "c": 2}


def _prediction_to_index(prediction: int | str) -> int | None:
    """Convert a prediction to an integer index if possible.

    Args:
        prediction: Model prediction as a numeric index or a choice label.

    Returns:
        Integer index for the prediction, or None if it cannot be parsed.
    """
    if isinstance(prediction, numbers.Integral):
        return int(prediction)
    if isinstance(prediction, str):
        cleaned = prediction.strip().lower()
        if cleaned in CHOICE_TO_INDEX:
            return CHOICE_TO_INDEX[cleaned]
        if cleaned.isdigit():
            return int(cleaned)
    return None


def _bias_adjusted_accuracy(acc: float, bias: float) -> float:
    """Accuracy minus a symmetric bias penalty (|bias|), clamped at zero.

    Keeps accuracy leading while subtracting bias directly.

    Args:
        acc: Raw accuracy value.
        bias: Signed bias value.

    Returns:
        Bias-adjusted accuracy clamped to zero.
    """
    penalty = abs(bias)
    return max(0.0, acc - penalty)


def _tally_ambig(
    pred_idx: int | None,
    stereo_idx: int | None,
    counter_idx: int | None,
    unknown_idx: int | None,
    counts: dict[str, int],
) -> None:
    """Update ambiguous-context counters in-place.

    Args:
        pred_idx: Parsed prediction index or None if unknown.
        stereo_idx: Index of the stereotype answer.
        counter_idx: Index of the counter-stereotype answer.
        unknown_idx: Index of the "unknown" answer, if available.
        counts: Mutable counter dictionary updated in-place.
    """
    counts["n_ambiguous"] += 1
    if pred_idx is None:
        return

    if pred_idx == unknown_idx:
        counts["n_correct_ambig"] += 1
        # If the model abstained (“unknown”), count accuracy but skip bias tally.
        if unknown_idx is not None:
            return

    if pred_idx == stereo_idx:
        counts["n_biased"] += 1
    elif pred_idx == counter_idx:
        counts["n_counterbiased"] += 1


class BiasMetric(Metric):
    """Bias and accuracy metrics for MBBQ (Neplenbroek et al., 2024)."""

    def __init__(
        self,
        name: str,
        pretty_name: str,
        bias_type: BiasType = "bias_adjusted_accuracy_ambig",
    ) -> None:
        """Initialise the bias metric.

        Context type
        - Ambiguous: correct answer should be "unknown/not enough information".

        Metrics
        - Ambiguous bias (bias_ambig): (stereotype picks − counter-stereotype picks) / n_ambiguous
        - Ambiguous accuracy (accuracy_ambig): correct "unknown" picks / n_ambiguous
        - Bias-adjusted accuracy: accuracy minus |bias|, clamped at zero.

        Notes:
        - "Unknown/not enough info" answers are ignored in bias numerators.
        - Returns NaN when the context type is absent.

        Args:
            name: Metric identifier.
            pretty_name: Human-readable metric name.
            bias_type: Metric variant to compute.
        """  # noqa: E501
        super().__init__(
            name=name,
            pretty_name=pretty_name,
            postprocessing_fn=lambda x: (x * 100, f"{x * 100:.1f}%"),
        )
        if bias_type not in VALID_BIAS_TYPES:
            raise ValueError(
                f"Unsupported bias_type {bias_type!r}; "
                f"choose one of {VALID_BIAS_TYPES!r}"
            )
        self.bias_type = bias_type

    def __call__(
        self,
        predictions: c.Sequence,
        references: c.Sequence,
        dataset: "Dataset",
        dataset_config: "DatasetConfig | None",
        benchmark_config: "BenchmarkConfig | None",
    ) -> float:
        """Compute the bias metric for the given predictions.

        Args:
            predictions:
                Model predictions, expected as choice indices or labels ("a"/"b"/"c").
            references:
                Unused for this metric, kept for interface compatibility.
            dataset:
                Dataset containing per-row metadata such as stereotype/counter indices.
            dataset_config:
                Unused for this metric, kept for interface compatibility.
            benchmark_config:
                Unused for this metric, kept for interface compatibility.

        Returns:
            The calculated metric score, or NaN when the relevant context type is
            absent.
        """
        counts = {
            "n_biased": 0,
            "n_counterbiased": 0,
            "n_ambiguous": 0,
            "n_correct_ambig": 0,
        }

        for pred, instance in zip(predictions, dataset):
            # Get all necessary meta information from the current instance
            stereo_idx = instance.get("stereo_idx")
            counter_idx = instance.get("counter_idx")
            unknown_idx = instance.get("unknown_idx")

            pred_idx = _prediction_to_index(prediction=pred)

            # Updates counts in-place for ambiguous-context tallies.
            _tally_ambig(
                pred_idx=pred_idx,
                stereo_idx=stereo_idx,
                counter_idx=counter_idx,
                unknown_idx=unknown_idx,
                counts=counts,
            )

        def bias_ambig() -> float:
            """Compute ambiguous-context bias for the current counts.

            Returns:
                Bias score, or NaN if there are no ambiguous instances.
            """
            if counts["n_ambiguous"] == 0:
                return float("nan")
            return (counts["n_biased"] - counts["n_counterbiased"]) / counts[
                "n_ambiguous"
            ]

        def accuracy_ambig() -> float:
            """Compute ambiguous-context accuracy for the current counts.

            Returns:
                Accuracy score, or NaN if there are no ambiguous instances.
            """
            if counts["n_ambiguous"] == 0:
                return float("nan")
            return counts["n_correct_ambig"] / counts["n_ambiguous"]

        def bias_adjusted_accuracy_ambig() -> float:
            """Compute bias-adjusted accuracy for ambiguous contexts.

            Returns:
                Bias-adjusted accuracy, or NaN if there are no ambiguous instances.
            """
            if counts["n_ambiguous"] == 0:
                return float("nan")
            acc = counts["n_correct_ambig"] / counts["n_ambiguous"]
            bias = (counts["n_biased"] - counts["n_counterbiased"]) / counts[
                "n_ambiguous"
            ]
            return _bias_adjusted_accuracy(acc=acc, bias=bias)

        metric_fns: dict[str, t.Callable[[], float]] = {
            "bias_ambig": bias_ambig,
            "accuracy_ambig": accuracy_ambig,
            "bias_adjusted_accuracy_ambig": bias_adjusted_accuracy_ambig,
        }

        return metric_fns[self.bias_type]()


bias_ambig_metric = BiasMetric(
    name="bias_ambig", pretty_name="Ambiguous context bias", bias_type="bias_ambig"
)

accuracy_ambig_metric = BiasMetric(
    name="accuracy_ambig",
    pretty_name="Ambiguous context accuracy",
    bias_type="accuracy_ambig",
)

bias_adjusted_accuracy_ambig_metric = BiasMetric(
    name="bias_adjusted_accuracy_ambig",
    pretty_name="Ambiguous bias-adjusted accuracy",
    bias_type="bias_adjusted_accuracy_ambig",
)



# /// script
# requires-python = ">=3.10,<4.0"
# dependencies = [
#     "pandas==2.2.0",
#     "requests==2.32.3",
# ]
# ///

"""Load the part-of-speech part of a Universal Dependencies treebank."""

import logging
import re
from collections import defaultdict
from functools import partial
from typing import Callable, DefaultDict
from urllib.parse import urlparse

import pandas as pd
import requests

logging.basicConfig(format="%(asctime)s ⋅ %(message)s", level=logging.INFO)
logger = logging.getLogger("load_ud_pos")


def load_dadt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Danish Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://github.com/UniversalDependencies/UD_Danish-DDT/raw/master/"
        "da_ddt-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_ptdt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Portuguese Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://github.com/UniversalDependencies/UD_Portuguese-Bosque/raw/master/"
        "pt_bosque-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(
        train_url=train_url,
        val_url=val_url,
        test_url=test_url,
        filter_source="CETEMPúblico",
    )


def load_fodt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Faroese Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://github.com/UniversalDependencies/UD_Faroese-FarPaHC/raw/master/"
        "fo_farpahc-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_isdt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Icelandic Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://github.com/UniversalDependencies/UD_Icelandic-Modern/raw/master/"
        "is_modern-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_nodt_nb_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Norwegian Bokmål Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://github.com/UniversalDependencies/UD_Norwegian-Bokmaal/raw/master/"
        "no_bokmaal-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_nodt_nn_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Norwegian Nynorsk Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://github.com/UniversalDependencies/UD_Norwegian-Nynorsk/raw/master/"
        "no_nynorsk-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_svdt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Swedish Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://github.com/UniversalDependencies/UD_Swedish-Talbanken/raw/master/"
        "sv_talbanken-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    # Define document processing function
    def process_document(doc: str) -> str:
        doc = (
            doc.replace(" s k", " s.k.")
            .replace("S k", "S.k.")
            .replace(" bl a", " bl.a.")
            .replace("Bl a", "Bl.a.")
            .replace(" t o m", " t.o.m.")
            .replace("T o m", "T.o.m.")
            .replace(" fr o m", " fr.o.m.")
            .replace("Fr o m", "Fr.o.m.")
            .replace(" o s v", " o.s.v.")
            .replace("O s v", "O.s.v.")
            .replace(" d v s", " d.v.s.")
            .replace("D v s", "D.v.s.")
            .replace(" m fl", " m.fl.")
            .replace("M fl", "M.fl.")
            .replace(" t ex", " t.ex.")
            .replace("T ex", "T.ex.")
            .replace(" f n", " f.n.")
            .replace("F n", "F.n.")
        )
        return doc

    return load_ud_pos(
        train_url=train_url,
        val_url=val_url,
        test_url=test_url,
        doc_process_fn=process_document,
    )


def load_dedt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the German Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_German-GSD/master/"
        "de_gsd-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_nldt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Dutch Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Dutch-Alpino/"
        "master/nl_alpino-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_endt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the English Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_English-GUM/"
        "master/en_gum-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_frdt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the French Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_French-GSD/"
        "master/fr_gsd-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_itdt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Italian Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Italian-ISDT/"
        "master/it_isdt-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_esdt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Spanish Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Spanish-AnCora"
        "/refs/heads/master/es_ancora-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_fidt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Finnish Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Finnish-TDT/"
        "master/fi_tdt-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_lvdt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Latvian Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Latvian-LVTB/"
        "master/lv_lvtb-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_etdt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Estonian Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Estonian-EDT/"
        "refs/heads/master/et_edt-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_pldt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Polish Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Polish-PDB/"
        "refs/heads/master/pl_pdb-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_ltdt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Lithuanian Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Lithuanian-ALKSNIS/"
        "master/lt_alksnis-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_csdt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Czech Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Czech-CAC/refs/"
        "heads/master/cs_cac-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_skdt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Slovak Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val`, and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Slovak-SNK/refs"
        "/heads/master/sk_snk-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_ukdt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Ukrainian Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/"
        "UD_Ukrainian-ParlaMint/refs/heads/master/uk_parlamint-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_eldt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Greek Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val`, and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Greek-GUD/refs"
        "/heads/master/el_gud-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_bgdt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Bulgarian Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Bulgarian-BTB/refs"
        "/heads/master/bg_btb-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_srdt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Serbian Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val`, and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Serbian-SET/refs"
        "/heads/master/sr_set-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_sldt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Slovene Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val`, and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Slovene-SSJ/refs"
        "/heads/master/sl_ssj-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_hrdt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Croatian Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val`, and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Croatian-SET/refs"
        "/heads/master/hr_set-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_hudt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Hungarian Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val`, and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Hungarian-Szeged/refs"
        "/heads/master/hu_szeged-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_rodt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Romanian Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val`, and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Romanian-RRT/refs"
        "/heads/master/ro_rrt-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_cadt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Catalan Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val`, and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Catalan-AnCora/refs"
        "/heads/master/ca_ancora-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_sqdt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Albanian Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val`, and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Albanian-STAF/refs"
        "/heads/master/sq_staf-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    df_dict = load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)

    # Load additional TSA test data and concatenate with the existing test split
    tsa_test_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Albanian-TSA/refs"
        "/heads/master/sq_tsa-ud-test.conllu"
    )
    tsa_test_lines = _load_file_or_url(url_or_path=tsa_test_url)
    tsa_test_df = _load_split(lines=tsa_test_lines)

    df_dict["test"] = pd.concat([df_dict["test"], tsa_test_df], ignore_index=True)

    return df_dict


def _load_file_or_url(url_or_path: str) -> list[str]:
    """Load a file from a URL or local path.

    Args:
        url_or_path: The URL or local path to load.

    Returns:
        The list of strings, one per line.
    """
    parsed = urlparse(url_or_path)
    if parsed.scheme.lower() in ("http", "https"):
        return requests.get(url_or_path).text.split("\n")
    else:
        with open(url_or_path, "r") as f:
            logger.warning(f"Loading data from local file: {url_or_path}")
            return [line.strip() for line in f.readlines()]


def _filter_token_range(data_dict: dict[str, list]) -> dict[str, list]:
    """Filter out tokens that belong to ranges in UD source files.

    Tokens that span more than one position are not supported by create_scala's
    prepare_df logic.

    Example files:

    - tests/test_scripts/test_create_scala/test_data/de_gsd-ud-train.conllu.adp_det
    - tests/test_scripts/test_create_scala/test_data/pl_pdb-ud-train.conllu.aux_clitic_*

    Args:
        data_dict: The input data dictionary.

    Returns:
        The filtered data dictionary. Its format is identical to the input.
    """
    output: DefaultDict[str, list] = defaultdict(list)

    range_start: int = 0
    range_end: int = 0
    # Track the index where we added the multi-word token so we can update its POS tag
    mwt_index: int | None = None

    # Pattern used to detect merged IDs
    merged_ids_pattern = re.compile(r"(\d+)-(\d+)", re.I)

    for i in range(len(data_dict["ids"])):
        match = merged_ids_pattern.match(data_dict["ids"][i])
        if match is not None:
            output["ids"].append(data_dict["ids"][i])
            output["tokens"].append(data_dict["tokens"][i])
            output["pos_tags"].append(data_dict["pos_tags"][i])
            range_start = int(match.group(1))
            range_end = int(match.group(2))
            # Remember where we added this multi-word token
            mwt_index = len(output["pos_tags"]) - 1
        else:
            token_id = int(data_dict["ids"][i].split(".")[0])
            if token_id >= range_start and token_id <= range_end:
                # If this is the first token in the range and we have a multi-word token
                # to update, copy the POS tag from this token to the multi-word token
                if mwt_index is not None and token_id == range_start:
                    output["pos_tags"][mwt_index] = data_dict["pos_tags"][i]
                    mwt_index = None  # Reset so we only update once per range
                # Skip token if in range
                continue
            else:
                output["ids"].append(data_dict["ids"][i])
                output["tokens"].append(data_dict["tokens"][i])
                output["pos_tags"].append(data_dict["pos_tags"][i])
                # Reset mwt_index when we move past the range
                mwt_index = None

    return output


def _load_split(
    *,
    lines: list[str],
    filter_source: str | None = None,
    doc_process_fn: Callable[[str], str] = lambda x: x,
) -> pd.DataFrame:
    """Load single split of the POS part of a Universal Dependencies treebank.

    Args:
        lines: The lines of the file to process.
        filter_source:
            If not `None`, only include entries with this source in the dataset.
        doc_process_fn:
            A function to apply to each document before parsing it.

    Returns:
        The dataframe for the given split.
    """
    # Initialise the records, data dictionary and document
    records: list[dict] = []
    data_dict: dict[str, list[list[int | str] | str]] = defaultdict(list)

    # Iterate over the data for the given split
    doc = ""
    source = ""
    for line_idx, line in enumerate(lines):
        # If we are at the first line of an entry then extract the document
        if line.startswith("# text = "):
            doc = re.sub("# text = ", "", line)

            # Process the document if needed
            doc = doc_process_fn(doc)

        elif line.startswith("# source = "):
            source = line.removeprefix("# source = ").strip()

        # Otherwise, if the line is a comment then ignore it
        elif line.startswith("#"):
            continue

        # Otherwise, if we have reached the end of an entry then store it to the
        # list of records and reset the data dictionary and document
        elif line == "" or line_idx == len(lines) - 1:
            if len(data_dict["tokens"]) > 0:
                if filter_source is None or filter_source in source:
                    merged_data_dict: dict[str, str | list[int | str]]
                    merged_data_dict = {**_filter_token_range(data_dict), "doc": doc}
                    records.append(merged_data_dict)
            data_dict = defaultdict(list)
            doc = ""
            source = ""

        # Otherwise we are in the middle of an entry which is not a comment, so
        # we extract the data from the line and store it in the data dictionary
        else:
            data_tup = line.split("\t")
            data_dict["ids"].append(data_tup[0])
            data_dict["tokens"].append(data_tup[1])
            data_dict["pos_tags"].append(data_tup[3])

    # Convert the records to a dataframe
    return pd.DataFrame.from_records(records)


def load_ud_pos(
    train_url: str,
    val_url: str,
    test_url: str,
    doc_process_fn: Callable[[str], str] = lambda x: x,
    filter_source: str | None = None,
) -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of a Universal Dependencies treebank.

    Args:
        train_url:
            The URL of the training data.
        val_url:
            The URL of the validation data.
        test_url:
            The URL of the test data.
        doc_process_fn:
            A function to apply to each document before parsing it.
        filter_source:
            If not `None`, only include entries with this source in the dataset.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    if filter_source is not None:
        logger.warning(
            f"Warning: Filtering dataset to include only entries with {filter_source=}"
        )

    # Load the lines of the splits
    train_lines = _load_file_or_url(url_or_path=train_url)
    val_lines = _load_file_or_url(url_or_path=val_url)
    test_lines = _load_file_or_url(url_or_path=test_url)

    # Load the splits
    split_loader = partial(
        _load_split, filter_source=filter_source, doc_process_fn=doc_process_fn
    )
    dfs = dict(
        train=split_loader(lines=train_lines),
        val=split_loader(lines=val_lines),
        test=split_loader(lines=test_lines),
    )

    # Return the dictionary of dataframes
    return dfs



"""All Icelandic dataset configurations used in EuroEval."""

from ..data_models import DatasetConfig
from ..languages import ICELANDIC
from ..tasks import COMMON_SENSE, EUROPEAN_VALUES, KNOW, LA, MCRC, NER, RC, SENT, SUMM

### Official datasets ###

HOTTER_AND_COLDER_SENTIMENT_CONFIG = DatasetConfig(
    name="hotter-and-colder-sentiment",
    pretty_name="Hotter and Colder Sentiment",
    source="EuroEval/hotter-and-colder-sentiment",
    task=SENT,
    languages=[ICELANDIC],
)

SCALA_IS_CONFIG = DatasetConfig(
    name="scala-is",
    pretty_name="ScaLA-is",
    source="EuroEval/scala-is",
    task=LA,
    languages=[ICELANDIC],
)

MIM_GOLD_NER_CONFIG = DatasetConfig(
    name="mim-gold-ner",
    pretty_name="MIM-GOLD-NER",
    source="EuroEval/mim-gold-ner-mini",
    task=NER,
    languages=[ICELANDIC],
)

NQII_CONFIG = DatasetConfig(
    name="nqii",
    pretty_name="NQiI",
    source="EuroEval/nqii-mini",
    task=RC,
    languages=[ICELANDIC],
)

RRN_CONFIG = DatasetConfig(
    name="rrn",
    pretty_name="RRN",
    source="EuroEval/rrn-mini",
    task=SUMM,
    languages=[ICELANDIC],
)

ICELANDIC_KNOWLEDGE_CONFIG = DatasetConfig(
    name="icelandic-knowledge",
    pretty_name="Icelandic Knowledge",
    source="EuroEval/icelandic-knowledge",
    task=KNOW,
    languages=[ICELANDIC],
)

WINOGRANDE_IS_CONFIG = DatasetConfig(
    name="winogrande-is",
    pretty_name="Winogrande-is",
    source="EuroEval/winogrande-is",
    task=COMMON_SENSE,
    languages=[ICELANDIC],
    _labels=["a", "b"],
)

VALEU_IS_CONFIG = DatasetConfig(
    name="valeu-is",
    pretty_name="VaLEU-is",
    source="EuroEval/european-values-is",
    task=EUROPEAN_VALUES,
    languages=[ICELANDIC],
    splits=["test"],
    bootstrap_samples=False,
    _instruction_prompt="{text}",
)


### Unofficial datasets ###

ICE_EC_CONFIG = DatasetConfig(
    name="ice-ec",
    pretty_name="ICE-EC",
    source="EuroEval/ice-ec",
    task=LA,
    languages=[ICELANDIC],
    unofficial=True,
)

ICE_EC_FULL_CONFIG = DatasetConfig(
    name="ice-ec-full",
    pretty_name="ICE-EC Full",
    source="EuroEval/ice-ec-full",
    task=LA,
    languages=[ICELANDIC],
    unofficial=True,
)

ICE_LINGUISTIC_CONFIG = DatasetConfig(
    name="ice-linguistic",
    pretty_name="IceLinguistic",
    source="EuroEval/ice-linguistic",
    task=LA,
    languages=[ICELANDIC],
    unofficial=True,
)

ICELANDIC_QA_CONFIG = DatasetConfig(
    name="icelandic-qa",
    pretty_name="Icelandic QA",
    source="EuroEval/icelandic-qa",
    task=RC,
    languages=[ICELANDIC],
    unofficial=True,
)

MMLU_IS_CONFIG = DatasetConfig(
    name="mmlu-is",
    pretty_name="MMLU-is",
    source="EuroEval/mmlu-is-mini",
    task=KNOW,
    languages=[ICELANDIC],
    unofficial=True,
)

ARC_IS_CONFIG = DatasetConfig(
    name="arc-is",
    pretty_name="ARC-is",
    source="EuroEval/arc-is-mini",
    task=KNOW,
    languages=[ICELANDIC],
    unofficial=True,
)

HELLASWAG_IS_CONFIG = DatasetConfig(
    name="hellaswag-is",
    pretty_name="HellaSwag-is",
    source="EuroEval/hellaswag-is-mini",
    task=COMMON_SENSE,
    languages=[ICELANDIC],
    unofficial=True,
)

BELEBELE_IS_CONFIG = DatasetConfig(
    name="belebele-is",
    pretty_name="Belebele-is",
    source="EuroEval/belebele-is-mini",
    task=MCRC,
    languages=[ICELANDIC],
    unofficial=True,
)

MULTI_WIKI_QA_IS_CONFIG = DatasetConfig(
    name="multi-wiki-qa-is",
    pretty_name="MultiWikiQA-is",
    source="EuroEval/multi-wiki-qa-is-mini",
    task=RC,
    languages=[ICELANDIC],
    unofficial=True,
)



"""All Faroese dataset configurations used in EuroEval."""

from ..data_models import DatasetConfig
from ..languages import FAROESE
from ..tasks import LA, NER, RC, SENT

### Official datasets ###

FOSENT_CONFIG = DatasetConfig(
    name="fosent",
    pretty_name="FoSent",
    source="EuroEval/fosent",
    task=SENT,
    languages=[FAROESE],
    _num_few_shot_examples=5,
)

SCALA_FO_CONFIG = DatasetConfig(
    name="scala-fo",
    pretty_name="ScaLA-fo",
    source="EuroEval/scala-fo",
    task=LA,
    languages=[FAROESE],
)

FONE_CONFIG = DatasetConfig(
    name="fone",
    pretty_name="FoNE",
    source="EuroEval/fone-mini",
    task=NER,
    languages=[FAROESE],
)

FOQA_CONFIG = DatasetConfig(
    name="foqa",
    pretty_name="FoQA",
    source="EuroEval/foqa",
    task=RC,
    languages=[FAROESE],
)


### Unofficial datasets ###

WIKIANN_FO_CONFIG = DatasetConfig(
    name="wikiann-fo",
    pretty_name="WikiANN-fo",
    source="EuroEval/wikiann-fo-mini",
    task=NER,
    languages=[FAROESE],
    unofficial=True,
)

MULTI_WIKI_QA_FO_CONFIG = DatasetConfig(
    name="multi-wiki-qa-fo",
    pretty_name="MultiWikiQA-fo",
    source="EuroEval/multi-wiki-qa-fo-mini",
    task=RC,
    languages=[FAROESE],
    unofficial=True,
)
